@@ -38,7 +38,7 @@ typedef struct Expand {
 #define IFS_WS		1	/* have seen IFS white-space */
 #define IFS_NWS		2	/* have seen IFS non-white-space */
 
-static	int	varsub ARGS((Expand *xp, char *sp, char *word, int *stypep));
+static	int	varsub ARGS((Expand *xp, char *sp, char *word, int *stypep, int *slenp));
 static	int	comsub ARGS((Expand *xp, char *cp));
 static	char   *trimsub ARGS((char *str, char *pat, int how));
 static	void	glob ARGS((char *cp, XPtrV *wp, int markdirs));
@@ -273,21 +273,23 @@ expand(cp, wp, f)
 				continue;
 			  case OSUBST: /* ${{#}var{:}[=+-?#%]word} */
 			  /* format is:
-			   *   OSUBST plain-variable-part \0
-			   *     compiled-word-part CSUBST
+			   *   OSUBST [{x] plain-variable-part \0
+			   *     compiled-word-part CSUBST [}x]
 			   * This is were all syntax checking gets done...
 			   */
 			  {
-				char *varname = sp;
+				char *varname = ++sp; /* skip the { or x (}) */
 				int stype;
+				int slen;
 
 				sp = strchr(sp, '\0') + 1; /* skip variable */
-				type = varsub(&x, varname, sp, &stype);
+				type = varsub(&x, varname, sp, &stype, &slen);
 				if (type < 0) {
 					char endc;
 					char *str, *end;
 
 					end = (char *) wdscan(sp, CSUBST);
+					/* ({) the } or x is already skipped */
 					endc = *end;
 					*end = EOS;
 					str = snptreef((char *) 0, 64, "%S",
@@ -315,12 +317,8 @@ expand(cp, wp, f)
 					st->var = x.var;
 					st->quote = quote;
 					/* skip qualifier(s) */
-					if (stype) {
-						sp += 2;
-						/* :[-+=?] or double [#%] */
-						if (stype & 0x80)
-							sp += 2;
-					}
+					if (stype)
+						sp += slen;
 					switch (stype & 0x7f) {
 					  case '#':
 					  case '%':
@@ -328,6 +326,12 @@ expand(cp, wp, f)
 						f = DOPAT | (f&DONTRUNCOMMAND)
 						    | DOTEMP_;
 						quote = 0;
+						/* Prepend open pattern (so |
+						 * in a trim will work as
+						 * expected)
+						 */
+						*dp++ = MAGIC;
+						*dp++ = '@' + 0x80;
 						break;
 					  case '=':
 						/* Enabling tilde expansion
@@ -343,13 +347,9 @@ expand(cp, wp, f)
 						 * sense though, since ~ is
 						 * a arithmetic operator.
 						 */
-#if !defined(__hppa) || __GNUC__ != 2	/* gcc 2.3.3 on hp-pa dies on this - ifdef goes away as soon as I get a new version of gcc.. */
 						if (!(x.var->flag & INTEGER))
 							f |= DOASNTILDE|DOTILDE;
 						f |= DOTEMP_;
-#else
-						f |= DOTEMP_|DOASNTILDE|DOTILDE;
-#endif
 						/* These will be done after the
 						 * value has been assigned.
 						 */
@@ -371,6 +371,7 @@ expand(cp, wp, f)
 				continue;
 			  }
 			  case CSUBST: /* only get here if expanding word */
+				sp++; /* ({) skip the } or x */
 				tilde_ok = 0;	/* in case of ${unset:-} */
 				*dp = '\0';
 				quote = st->quote;
@@ -380,6 +381,8 @@ expand(cp, wp, f)
 				switch (st->stype&0x7f) {
 				  case '#':
 				  case '%':
+					/* Append end-pattern */
+					*dp++ = MAGIC; *dp++ = ')'; *dp = '\0';
 					dp = Xrestpos(ds, dp, st->base);
 					/* Must use st->var since calling
 					 * global would break things
@@ -679,15 +682,17 @@ expand(cp, wp, f)
  * Prepare to generate the string returned by ${} substitution.
  */
 static int
-varsub(xp, sp, word, stypep)
+varsub(xp, sp, word, stypep, slenp)
 	Expand *xp;
 	char *sp;
 	char *word;
-	int *stypep;
+	int *stypep;	/* becomes qualifier type */
+	int *slenp;	/* " " len (=, :=, etc.) valid iff *stypep != 0 */
 {
 	int c;
 	int state;	/* next state: XBASE, XARG, XSUB, XNULLSUB */
 	int stype;	/* substitution type */
+	int slen;
 	char *p;
 	struct tbl *vp;
 
@@ -729,29 +734,34 @@ varsub(xp, sp, word, stypep)
 
 	/* Check for qualifiers in word part */
 	stype = 0;
-	c = *word == CHAR ? word[1] : 0;
+	c = word[slen = 0] == CHAR ? word[1] : 0;
 	if (c == ':') {
+		slen += 2;
 		stype = 0x80;
-		c = word[2] == CHAR ? word[3] : 0;
+		c = word[slen + 0] == CHAR ? word[slen + 1] : 0;
 	}
-	if (ctype(c, C_SUBOP1))
+	if (ctype(c, C_SUBOP1)) {
+		slen += 2;
 		stype |= c;
-	else if (stype)	/* :, :# or :% is not ok */
-		return -1;
-	else if (ctype(c, C_SUBOP2)) {
+	} else if (ctype(c, C_SUBOP2)) { /* Note: ksh88 allows :%, :%%, etc */
+		slen += 2;
 		stype = c;
-		if (word[2] == CHAR && c == word[3])
+		if (word[slen + 0] == CHAR && c == word[slen + 1]) {
 			stype |= 0x80;
-	}
+			slen += 2;
+		}
+	} else if (stype)	/* : is not ok */
+		return -1;
 	if (!stype && *word != CSUBST)
 		return -1;
 	*stypep = stype;
+	*slenp = slen;
 
 	c = sp[0];
 	if (c == '*' || c == '@') {
 		switch (stype & 0x7f) {
 		  case '=':	/* can't assign to a vector */
-		  case '%':	/* can't trim a vector */
+		  case '%':	/* can't trim a vector (yet) */
 		  case '#':
 			return -1;
 		}
@@ -770,7 +780,7 @@ varsub(xp, sp, word, stypep)
 
 			switch (stype & 0x7f) {
 			  case '=':	/* can't assign to a vector */
-			  case '%':	/* can't trim a vector */
+			  case '%':	/* can't trim a vector (yet) */
 			  case '#':
 				return -1;
 			}
@@ -1166,11 +1176,12 @@ debunk(dp, sp)
 		memcpy(dp, sp, s - sp);
 		for (d = dp + (s - sp); *s; s++)
 			if (!ISMAGIC(*s) || !(*++s & 0x80)
-			    || !strchr("*+?@!", *s & 0x7f))
+			    || !strchr("*+?@! ", *s & 0x7f))
 				*d++ = *s;
 			else {
 				/* extended pattern operators: *+?@! */
-				*d++ = *s & 0x7f;
+				if ((*s & 0x7f) != ' ')
+					*d++ = *s & 0x7f;
 				*d++ = '(';
 			}
 		*d = '\0';
