@@ -5,14 +5,61 @@
 #include "sh.h"
 #include <ctype.h>
 
+
+/* Structure to keep track of the lexing state and the various pieces of info
+ * needed for each particular state.
+ */
+typedef struct lex_state Lex_state;
+struct lex_state {
+	int ls_state;
+	union {
+	    /* $(...) */
+	    struct scsparen_info {
+		    int nparen;		/* count open parenthesis */
+		    int csstate; /* XXX remove */
+#define ls_scsparen ls_info.u_scsparen
+	    } u_scsparen;
+
+	    /* $((...)) */
+	    struct sasparen_info {
+		    int nparen;		/* count open parenthesis */
+		    int start;		/* marks start of $(( in output str */
+#define ls_sasparen ls_info.u_sasparen
+	    } u_sasparen;
+
+	    /* ((...)) */
+	    struct sletparen_info {
+		    int nparen;		/* count open parenthesis */
+#define ls_sletparen ls_info.u_sletparen
+	    } u_sletparen;
+
+	    /* `...` */
+	    struct sbquote_info {
+		    int indquotes;	/* true if in double quotes: "`...`" */
+#define ls_sbquote ls_info.u_sbquote
+	    } u_sbquote;
+
+	    Lex_state *base;		/* used to point to next state block */
+	} ls_info;
+};
+
+typedef struct State_info State_info;
+struct State_info {
+	Lex_state	*base;
+	Lex_state	*end;
+};
+
+
 static void	readhere ARGS((struct ioword *iop));
 static int	getsc__ ARGS((void));
 static void	getsc_line ARGS((Source *s));
+static int	getsc_bn ARGS((void));
 static char	*get_brace_var ARGS((XString *wsp, char *wp));
 static int	arraysub ARGS((char **strp));
 static const char *ungetsc ARGS((int c));
-static int	getsc_bn ARGS((void));
 static void	gethere ARGS((void));
+static Lex_state *push_state_ ARGS((State_info *si, Lex_state *old_end));
+static Lex_state *pop_state_ ARGS((State_info *si, Lex_state *old_end));
 
 static int backslash_skip;
 static int ignore_backslash_newline;
@@ -23,6 +70,21 @@ static int ignore_backslash_newline;
 /* optimized getsc__() */
 #define	getsc_()	((*source->str != '\0') ? *source->str++ : getsc__())
 
+#define STATE_BSIZE	32
+
+#define PUSH_STATE(s)	do { \
+			    if (++statep == state_info.end) \
+				statep = push_state_(&state_info, statep); \
+			    state = statep->ls_state = (s); \
+			} while (0)
+
+#define POP_STATE()	do { \
+			    if (--statep == state_info.base) \
+				statep = pop_state_(&state_info, statep); \
+			    state = statep->ls_state; \
+			} while (0)
+
+
 
 /*
  * Lexical analyzer
@@ -36,36 +98,38 @@ int
 yylex(cf)
 	int cf;
 {
+	Lex_state states[STATE_BSIZE], *statep;
+	State_info state_info;
 	register int c, state;
-	char states [64], *statep = states; /* XXX overflow check */
 	XString ws;		/* expandable output word */
 	register char *wp;	/* output word pointer */
-	register char *sp, *dp;
-	char UNINITIALIZED(*ddparen_start);
-	int istate;
-	int UNINITIALIZED(c2);
-	int UNINITIALIZED(nparen), UNINITIALIZED(csstate);
-	int UNINITIALIZED(ndparen);
-	int UNINITIALIZED(indquotes);
+	char *sp, *dp;
+	int c2;
 
 
   Again:
+	states[0].ls_state = -1;
+	states[0].ls_info.base = (Lex_state *) 0;
+	statep = &states[1];
+	state_info.base = states;
+	state_info.end = &states[STATE_BSIZE];
+
 	Xinit(ws, wp, 64, ATEMP);
 
 	backslash_skip = 0;
 	ignore_backslash_newline = 0;
 
 	if (cf&ONEWORD)
-		istate = SWORD;
+		state = SWORD;
 #ifdef KSH
 	else if (cf&LETEXPR) {
 		*wp++ = OQUOTE;	 /* enclose arguments in (double) quotes */
-		istate = SDPAREN;	
-		ndparen = 0;
+		state = SLETPAREN;	
+		statep->ls_sletparen.nparen = 0;
 	}
 #endif /* KSH */
 	else {		/* normal lexing */
-		istate = (cf & HEREDELIM) ? SHEREDELIM : SBASE;
+		state = (cf & HEREDELIM) ? SHEREDELIM : SBASE;
 		while ((c = getsc()) == ' ' || c == '\t')
 			;
 		if (c == '#') {
@@ -85,10 +149,13 @@ yylex(cf)
 			cf |= ALIAS;
 	}
 
+	/* Initial state: one of SBASE SHEREDELIM SWORD SASPAREN */
+	statep->ls_state = state;
+
 	/* collect non-special or quoted characters to form word */
-	for (*statep = state = istate;
-	     !((c = getsc()) == 0 || ((state == SBASE || state == SHEREDELIM)
-				      && ctype(c, C_LEX1))); )
+	while (!((c = getsc()) == 0
+		 || ((state == SBASE || state == SHEREDELIM)
+		     && ctype(c, C_LEX1))))
 	{
 		Xcheck(ws, wp);
 		switch (state) {
@@ -134,7 +201,7 @@ yylex(cf)
 				if (c2 == '(' /*)*/ ) {
 					*wp++ = OPAT;
 					*wp++ = c;
-					*++statep = state = SPATTERN;
+					PUSH_STATE(SPATTERN);
 					break;
 				}
 				ungetsc(c2);
@@ -155,13 +222,13 @@ yylex(cf)
 					*wp++ = QCHAR, *wp++ = c;
 				break;
 			  case '\'':
-				*++statep = state = SSQUOTE;
 				*wp++ = OQUOTE;
 				ignore_backslash_newline++;
+				PUSH_STATE(SSQUOTE);
 				break;
 			  case '"':
-				*++statep = state = SDQUOTE;
 				*wp++ = OQUOTE;
+				PUSH_STATE(SDQUOTE);
 				break;
 			  default:
 				goto Subst;
@@ -191,39 +258,41 @@ yylex(cf)
 				if (c == '(') /*)*/ {
 					c = getsc();
 					if (c == '(') /*)*/ {
-						*++statep = state = SDDPAREN;
-						nparen = 2;
-						ddparen_start = wp;
+						PUSH_STATE(SASPAREN);
+						statep->ls_sasparen.nparen = 2;
+						statep->ls_sasparen.start =
+							Xsavepos(ws, wp);
 						*wp++ = EXPRSUB;
 					} else {
 						ungetsc(c);
-						*++statep = state = SPAREN;
-						nparen = 1;
-						csstate = 0;
+						PUSH_STATE(SCSPAREN);
+						statep->ls_scsparen.nparen = 1;
+						statep->ls_scsparen.csstate = 0;
 						*wp++ = COMSUB;
 					}
 				} else if (c == '{') /*}*/ {
 					*wp++ = OSUBST;
+					*wp++ = '{'; /*}*/
 					wp = get_brace_var(&ws, wp);
-					/* If this is a trim operation,
-					 * wrap @(...) around the pattern
-					 * (allows easy handling of ${a#b|c})
-					 */
 					c = getsc();
-					if (c == '#' || c == '%') {
+					/* allow :# and :% (ksh88 compat) */
+					if (c == ':') {
 						*wp++ = CHAR, *wp++ = c;
-						if ((c2 = getsc()) == c)
-							*wp++ = CHAR, *wp++ = c;
-						else
-							ungetsc(c2);
-						*wp++ = OPAT, *wp++ = '@';
-						*++statep = state = STBRACE;
+						c = getsc();
+					}
+					/* If this is a trim operation,
+					 * treat (,|,) specially in STBRACE.
+					 */
+					if (c == '#' || c == '%') {
+						ungetsc(c);
+						PUSH_STATE(STBRACE);
 					} else {
 						ungetsc(c);
-						*++statep = state = SBRACE;
+						PUSH_STATE(SBRACE);
 					}
 				} else if (ctype(c, C_ALPHA)) {
 					*wp++ = OSUBST;
+					*wp++ = 'X';
 					do {
 						Xcheck(ws, wp);
 						*wp++ = c;
@@ -231,30 +300,62 @@ yylex(cf)
 					} while (ctype(c, C_ALPHA|C_DIGIT));
 					*wp++ = '\0';
 					*wp++ = CSUBST;
+					*wp++ = 'X';
 					ungetsc(c);
 				} else if (ctype(c, C_DIGIT|C_VAR1)) {
 					Xcheck(ws, wp);
 					*wp++ = OSUBST;
+					*wp++ = 'X';
 					*wp++ = c;
 					*wp++ = '\0';
 					*wp++ = CSUBST;
+					*wp++ = 'X';
 				} else {
 					*wp++ = CHAR, *wp++ = '$';
 					ungetsc(c);
 				}
 				break;
 			  case '`':
-				*++statep = state = SBQUOTE;
+				PUSH_STATE(SBQUOTE);
 				*wp++ = COMSUB;
 				/* Need to know if we are inside double quotes
 				 * since sh/at&t-ksh translate the \" to " in
 				 * "`..\"..`".
+				 * This is not done in posix mode (section
+				 * 3.2.3, Double Quotes: "The backquote shall
+				 * retain its special meaning introducing the
+				 * other form of command substitution (see
+				 * 3.6.3). The portion of the quoted string
+				 * from the initial backquote and the
+				 * characters up to the next backquote that
+				 * is not preceded by a backslash (having
+				 * escape characters removed) defines that
+				 * command whose output replaces `...` when
+				 * the word is expanded."
+				 * Section 3.6.3, Command Substitution:
+				 * "Within the backquoted style of command
+				 * substitution, backslash shall retain its
+				 * literal meaning, except when followed by
+				 * $ ` \.").
 				 */
-				indquotes = 0;
-				if (!Flag(FPOSIX))
-					for (sp = statep; sp > states; --sp)
-						if (*sp == SDQUOTE)
-							indquotes = 1;
+				statep->ls_sbquote.indquotes = 0;
+				if (!Flag(FPOSIX)) {
+					Lex_state *s = statep;
+					Lex_state *base = state_info.base;
+					while (1) {
+						for (; s != base; s--) {
+							if (s->ls_state == SDQUOTE) {
+								statep->ls_sbquote.indquotes = 1;
+								break;
+							}
+						}
+						if (s != base)
+							break;
+						if (!(s = s->ls_info.base))
+							break;
+						base = s-- - STATE_BSIZE;
+					}
+				}
 				break;
 			  default:
 				*wp++ = CHAR, *wp++ = c;
@@ -263,7 +364,7 @@ yylex(cf)
 
 		  case SSQUOTE:
 			if (c == '\'') {
-				state = *--statep;
+				POP_STATE();
 				*wp++ = CQUOTE;
 				ignore_backslash_newline--;
 			} else
@@ -272,35 +373,35 @@ yylex(cf)
 
 		  case SDQUOTE:
 			if (c == '"') {
-				state = *--statep;
+				POP_STATE();
 				*wp++ = CQUOTE;
 			} else
 				goto Subst;
 			break;
 
-		  case SPAREN: /* $( .. ) */
+		  case SCSPAREN: /* $( .. ) */
 			/* todo: deal with $(...) quoting properly
 			 * kludge to partly fake quoting inside $(..): doesn't
 			 * really work because nested $(..) or ${..} inside
 			 * double quotes aren't dealt with.
 			 */
-			switch (csstate) {
+			switch (statep->ls_scsparen.csstate) {
 			  case 0: /* normal */
 				switch (c) {
 				  case '(':
-					nparen++;
+					statep->ls_scsparen.nparen++;
 					break;
 				  case ')':
-					nparen--;
+					statep->ls_scsparen.nparen--;
 					break;
 				  case '\\':
-					csstate = 1;
+					statep->ls_scsparen.csstate = 1;
 					break;
 				  case '"':
-					csstate = 2;
+					statep->ls_scsparen.csstate = 2;
 					break;
 				  case '\'':
-					csstate = 4;
+					statep->ls_scsparen.csstate = 4;
 					ignore_backslash_newline++;
 					break;
 				}
@@ -308,58 +409,63 @@ yylex(cf)
 
 			  case 1: /* backslash in normal mode */
 			  case 3: /* backslash in double quotes */
-				--csstate;
+				--statep->ls_scsparen.csstate;
 				break;
 
 			  case 2: /* double quotes */
 				if (c == '"')
-					csstate = 0;
+					statep->ls_scsparen.csstate = 0;
 				else if (c == '\\')
-					csstate = 3;
+					statep->ls_scsparen.csstate = 3;
 				break;
 
 			  case 4: /* single quotes */
 				if (c == '\'') {
-					csstate = 0;
+					statep->ls_scsparen.csstate = 0;
 					ignore_backslash_newline--;
 				}
 				break;
 			}
-			if (nparen == 0) {
-				state = *--statep;
+			if (statep->ls_scsparen.nparen == 0) {
+				POP_STATE();
 				*wp++ = 0; /* end of COMSUB */
 			} else
 				*wp++ = c;
 			break;
 
-		  case SDDPAREN: /* $(( .. )) */
+		  case SASPAREN: /* $(( .. )) */
 			/* todo: deal with $((...); (...)) properly */
 			/* XXX should nest using existing state machine
 			 *     (embed "..", $(...), etc.) */
 			if (c == '(')
-				nparen++;
+				statep->ls_sasparen.nparen++;
 			else if (c == ')') {
-				nparen--;
-				if (nparen == 1) {
+				statep->ls_sasparen.nparen--;
+				if (statep->ls_sasparen.nparen == 1) {
 					/*(*/
 					if ((c2 = getsc()) == ')') {
-						state = *--statep;
+						POP_STATE();
 						*wp++ = 0; /* end of EXPRSUB */
 						break;
 					} else {
+						char *s;
+
 						ungetsc(c2);
 						/* mismatched parenthesis -
 						 * assume we were really
 						 * parsing a $(..) expression
 						 */
-						memmove(ddparen_start + 1,
-							ddparen_start,
-							wp - ddparen_start);
-						*ddparen_start++ = COMSUB;
-						*ddparen_start = '('; /*)*/
+						s = Xrestpos(ws, wp,
+						     statep->ls_sasparen.start);
+						memmove(s + 1, s, wp - s);
+						*s++ = COMSUB;
+						*s = '('; /*)*/
 						wp++;
-						csstate = 0;
-						*statep = state = SPAREN;
+						statep->ls_scsparen.nparen = 1;
+						statep->ls_scsparen.csstate = 0;
+						state = statep->ls_state
+							= SCSPAREN;
+						
 					}
 				}
 			}
@@ -369,23 +475,26 @@ yylex(cf)
 		  case SBRACE:
 			/*{*/
 			if (c == '}') {
-				state = *--statep;
+				POP_STATE();
 				*wp++ = CSUBST;
+				*wp++ = /*{*/ '}';
 			} else
 				goto Sbase1;
 			break;
 
 		  case STBRACE:
-			/* same as SBRACE, except | is saved as SPAT and
-			 * CPAT is added at the end.
-			 */
+			/* Same as SBRACE, except (,|,) treated specially */
 			/*{*/
 			if (c == '}') {
-				state = *--statep;
-				*wp++ = CPAT;
+				POP_STATE();
 				*wp++ = CSUBST;
+				*wp++ = /*{*/ '}';
 			} else if (c == '|') {
 				*wp++ = SPAT;
+			} else if (c == '(') {
+				*wp++ = OPAT;
+				*wp++ = ' ';	/* simile for @ */
+				PUSH_STATE(SPATTERN);
 			} else
 				goto Sbase1;
 			break;
@@ -393,7 +502,7 @@ yylex(cf)
 		  case SBQUOTE:
 			if (c == '`') {
 				*wp++ = 0;
-				state = *--statep;
+				POP_STATE();
 			} else if (c == '\\') {
 				switch (c = getsc()) {
 				  case '\\':
@@ -401,7 +510,7 @@ yylex(cf)
 					*wp++ = c;
 					break;
 				  case '"':
-					if (indquotes) {
+					if (statep->ls_sbquote.indquotes) {
 						*wp++ = c;
 						break;
 					}
@@ -421,11 +530,11 @@ yylex(cf)
 			goto Subst;
 
 #ifdef KSH
-		  case SDPAREN:	/* LETEXPR: (( ... )) */
+		  case SLETPAREN:	/* LETEXPR: (( ... )) */
 			/*(*/
 			if (c == ')') {
-				if (ndparen > 0)
-				    --ndparen;
+				if (statep->ls_sletparen.nparen > 0)
+				    --statep->ls_sletparen.nparen;
 				/*(*/
 				else if ((c2 = getsc()) == ')') {
 					c = 0;
@@ -438,7 +547,7 @@ yylex(cf)
 				 * are lost, but at&t ksh doesn't count them
 				 * either
 				 */
-				++ndparen;
+				++statep->ls_sletparen.nparen;
 			goto Sbase2;
 #endif /* KSH */
 
@@ -458,11 +567,11 @@ yylex(cf)
 					*wp++ = c;
 				}
 			} else if (c == '\'') {
-				*++statep = state = SSQUOTE;
+				PUSH_STATE(SSQUOTE);
 				*wp++ = OQUOTE;
 				ignore_backslash_newline++;
 			} else if (c == '"') {
-				state = SHEREDQUOTE;
+				state = statep->ls_state = SHEREDQUOTE;
 				*wp++ = OQUOTE;
 			} else {
 				*wp++ = CHAR;
@@ -473,7 +582,7 @@ yylex(cf)
 		  case SHEREDQUOTE:	/* " in <<,<<- delimiter */
 			if (c == '"') {
 				*wp++ = CQUOTE;
-				state = SHEREDELIM;
+				state = statep->ls_state = SHEREDELIM;
 			} else {
 				if (c == '\\') {
 					switch (c = getsc()) {
@@ -496,33 +605,69 @@ yylex(cf)
 		  case SPATTERN:	/* in *(...|...) pattern (*+?@!) */
 			if ( /*(*/ c == ')') {
 				*wp++ = CPAT;
-				state = *--statep;
-			} else if (c == '|')
+				POP_STATE();
+			} else if (c == '|') {
 				*wp++ = SPAT;
-			else
+			} else if (c == '(') {
+				*wp++ = OPAT;
+				*wp++ = ' ';	/* simile for @ */
+				PUSH_STATE(SPATTERN);
+			} else
 				goto Sbase1;
 			break;
 		}
 	}
 Done:
 	Xcheck(ws, wp);
-	if (state != istate)
+	if (statep != &states[1])
+		/* XXX figure out what is missing */
 		yyerror("no closing quote\n");
 
 	/* This done to avoid tests for SHEREDELIM wherever SBASE tested */
 	if (state == SHEREDELIM)
 		state = SBASE;
 
-	if ((c == '<' || c == '>') && state == SBASE) {
-		char *cp = Xstring(ws, wp);
-		if (Xlength(ws, wp) == 2 && cp[0] == CHAR && digit(cp[1])) {
-			wp = cp; /* throw away word */
-			c2/*unit*/ = cp[1] - '0';
-		} else
-			c2/*unit*/ = c == '>'; /* 0 for <, 1 for > */
+	dp = Xstring(ws, wp);
+	if ((c == '<' || c == '>') && state == SBASE
+	    && ((c2 = Xlength(ws, wp)) == 0
+	        || (c2 == 2 && dp[0] == CHAR && digit(dp[1]))))
+	{
+		struct ioword *iop =
+				(struct ioword *) alloc(sizeof(*iop), ATEMP);
+
+		if (c2 == 2)
+			iop->unit = dp[1] - '0';
+		else
+			iop->unit = c == '>'; /* 0 for <, 1 for > */
+
+		c2 = getsc();
+		/* <<, >>, <> are ok, >< is not */
+		if (c == c2 || (c == '<' && c2 == '>')) {
+			iop->flag = c == c2 ?
+				  (c == '>' ? IOCAT : IOHERE) : IORDWR;
+			if (iop->flag == IOHERE)
+				if ((c2 = getsc()) == '-')
+					iop->flag |= IOSKIP;
+				else
+					ungetsc(c2);
+		} else if (c2 == '&')
+			iop->flag = IODUP | (c == '<' ? IORDUP : 0);
+		else {
+			iop->flag = c == '>' ? IOWRITE : IOREAD;
+			if (c == '>' && c2 == '|')
+				iop->flag |= IOCLOB;
+			else
+				ungetsc(c2);
+		}
+
+		iop->name = (char *) 0;
+		iop->delim = (char *) 0;
+		Xfree(ws, wp);	/* free word */
+		yylval.iop = iop;
+		return REDIR;
 	}
 
-	if (wp == Xstring(ws, wp) && state == SBASE) {
+	if (wp == dp && state == SBASE) {
 		Xfree(ws, wp);	/* free word */
 		/* no word, process LEX1 character */
 		switch (c) {
@@ -545,38 +690,6 @@ Done:
 				ungetsc(c2);
 			return c;
 
-		  case '>':
-		  case '<': {
-			register struct ioword *iop;
-
-			iop = (struct ioword *) alloc(sizeof(*iop), ATEMP);
-			iop->unit = c2/*unit*/;
-
-			c2 = getsc();
-			/* <<, >>, <> are ok, >< is not */
-			if (c == c2 || (c == '<' && c2 == '>')) {
-				iop->flag = c == c2 ?
-					  (c == '>' ? IOCAT : IOHERE) : IORDWR;
-				if (iop->flag == IOHERE)
-					if ((c2 = getsc()) == '-')
-						iop->flag |= IOSKIP;
-					else
-						ungetsc(c2);
-			} else if (c2 == '&')
-				iop->flag = IODUP | (c == '<' ? IORDUP : 0);
-			else {
-				iop->flag = c == '>' ? IOWRITE : IOREAD;
-				if (c == '>' && c2 == '|')
-					iop->flag |= IOCLOB;
-				else
-					ungetsc(c2);
-			}
-
-			iop->name = (char *) 0;
-			iop->delim = (char *) 0;
-			yylval.iop = iop;
-			return REDIR;
-		    }
 		  case '\n':
 			gethere();
 			if (cf & CONTIN)
@@ -586,6 +699,7 @@ Done:
 		  case '(':  /*)*/
 #ifdef KSH
 			if ((c2 = getsc()) == '(') /*)*/
+				/* XXX need to handle ((...); (...)) */
 				c = MDPAREN;
 			else
 				ungetsc(c2);
@@ -601,7 +715,7 @@ Done:
 	yylval.cp = Xclose(ws, wp);
 	if (state == SWORD
 #ifdef KSH
-		|| state == SDPAREN
+		|| state == SLETPAREN
 #endif /* KSH */
 		)	/* ONEWORD? */
 		return LWORD;
@@ -750,7 +864,6 @@ yyerror(fmt, va_alist)
 {
 	va_list va;
 
-	yynerrs++;
 	/* pop aliases and re-reads */
 	while (source->type == SALIAS || source->type == SREREAD)
 		source = source->next;
@@ -1273,3 +1386,31 @@ getsc_bn ARGS((void))
 		return c;
 	}
 }
+
+static Lex_state *
+push_state_(si, old_end)
+	State_info *si;
+	Lex_state *old_end;
+{
+	Lex_state	*new = alloc(sizeof(Lex_state) * STATE_BSIZE, ATEMP);
+
+	new[0].ls_info.base = old_end;
+	si->base = &new[0];
+	si->end = &new[STATE_BSIZE];
+	return &new[1];
+}
+
+static Lex_state *
+pop_state_(si, old_end)
+	State_info *si;
+	Lex_state *old_end;
+{
+	Lex_state *old_base = si->base;
+
+	si->base = old_end->ls_info.base - STATE_BSIZE;
+	si->end = old_end->ls_info.base;
+
+	afree(old_base, ATEMP);
+
+	return si->base + STATE_BSIZE - 1;;
+}
