@@ -34,6 +34,7 @@ newblock()
 	static char *const empty[] = {null};
 
 	l = (struct block *) alloc(sizeof(struct block), ATEMP);
+	l->flags = 0;
 	ainit(&l->area);
 	if (!e->loc) {
 		l->argc = 0;
@@ -66,6 +67,8 @@ popblock()
 				setspec(vq);
 			else
 				unsetspec(vq);
+	if (l->flags & BF_DOGETOPTS)
+		user_opt = l->getopts_state;
 	afreeall(&l->area);
 	afree(l, ATEMP);
 }
@@ -881,6 +884,11 @@ getspec(vp)
 		vp->flag |= SPECIAL;
 		break;
 #endif /* HISTORY */
+	  case V_OPTIND:
+		vp->flag &= ~SPECIAL;
+		setint(vp, (long) user_opt.uoptind);
+		vp->flag |= SPECIAL;
+		break;
 	}
 }
 
@@ -892,7 +900,9 @@ setspec(vp)
 
 	switch (special(vp->name)) {
 	  case V_PATH:
-		path = str_val(vp);
+		if (path)
+			afree(path, APERM);
+		path = str_save(str_val(vp), APERM);
 		flushcom(1);	/* clear tracked aliases */
 		break;
 	  case V_IFS:
@@ -900,7 +910,9 @@ setspec(vp)
 		ifs0 = *s;
 		break;
 	  case V_OPTIND:
+		vp->flag &= ~SPECIAL;
 		getopts_reset((int) intval(vp));
+		vp->flag |= SPECIAL;
 		break;
 	  case V_POSIXLY_CORRECT:
 		change_flag(FPOSIX, OF_SPECIAL, 1);
@@ -979,7 +991,9 @@ unsetspec(vp)
 {
 	switch (special(vp->name)) {
 	  case V_PATH:
-		path = def_path;
+		if (path)
+			afree(path, APERM);
+		path = str_save(def_path, APERM);
 		flushcom(1);	/* clear tracked aliases */
 		break;
 	  case V_IFS:
